unit CParserData;

interface

uses
  Classes;

type
  TParserOption = (
    poKeep,             // keep original declaration (commented out)
    poExternalsym,      // add $EXTERNALSYM to declarations
    poPretty,           // add pretty names
    poConstants,        // convert enums to constants
    poVariant,          // convert union to (named) variant record
    poVarParams,        // prefer var parameters instead of pointer types
    poGroup,            // group parameters with same type
    poBorland,          // generate Borland style function declaration
    poDWords,           // generate DWORD macro function parameters
    poInlines           // Generate inline directive for macro functions
  );
  TParserOptions = set of TParserOption;

  PKnownType = ^TKnownType;
  TKnownType = record
  public
    CType: string;
    DelphiType: string;
    PtrLevel: Integer;
  end;

  TCParserData = class
  private
    class var
      FMacros: TStringList;
      FTypes: TStringList;
      FOptions: TParserOptions;
  public
    class procedure Initialize;
    class procedure Finalize;

    class procedure SetMacros(const Macros: TStrings); static;
    class procedure AddMacros(const Macros: array of string); static;
    class procedure AddMacro(const AMacro: string); static;

    class procedure SetTypes(const Types: TStrings); static;
    class procedure AddTypes(const Types: array of string); static;
    class procedure AddType(const AType: string); static;

    class procedure SetOptions(const Value: TParserOptions); static;

    class function StringToType(S: string): TKnownType; static;

    class property Macros: TStringList read FMacros;
    class property Types: TStringList read FTypes;
    class property Options: TParserOptions read FOptions write SetOptions;
  end;

const
  CTypes: array[1..17] of string = (
    'bool=Boolean',
    'Byte*=PByte',
    'Char*=PAnsiChar',
    'char=AnsiChar',
    'Double*=PDouble',
    'GUID=TGUID',
    'HANDLE=THandle',
    'Integer*=PInteger',
    'LPBYTE=PByte',
    'LPVOID=Pointer',
    'PBYTE=PByte',
    'PWSTR=PWideChar',
    'void**=PPointer',
    'void*=Pointer',
    'wchar_t*=PWideChar',
    'wchar_t=WideChar',
    ''
  );

  CMacros: array[1..71] of string = (
    '__in bool=#',
    '__deref_opt_out=#',
    '__deref_out_opt=#',
    '__deref_out=#',
    '__in_opt=#',
    '__in_z=#',
    '__in=#',
    '__inout_opt=#',
    '__inout=#',
    '__out_opt=#',
    '__out_z=#',
    '__out=#',
    '__RPC__inout_opt=#',
    '__RPC__out=#',
    '__unaligned=#',
    '_FAR=#',
    '_SIZE_T=unsigned',
    '_WIN_T=wchar_t',
    'APIENTRY=__stdcall',
    'APIPRIVATE=__stdcall',
    'BROWSEUIAPI=STDAPI',
    'BYTE=unsigned char',
    'CALLBACK=__stdcall',
    'CDECL=__cdecl',
    'CHAR=char',
    'DECLSPEC_EXPORT=__declspec(dllexport)',
    'DECLSPEC_IMPORT=__declspec(dllimport)',
    'DWMAPI=STDAPI',
    'EXTERN_C=extern "C"',
    'FAR=#',
    'HRESULT=HResult',
    'IN=#',
    'LPCSTR=PAnsiChar',
    'LPCTSTR=PChar',
    'LPCWSTR=PWideChar',
    'LPDWORD=DWORD *',
    'LPSTR=PAnsiChar',
    'LPTSTR=PChar',
    'LPVOID=void *',
    'LPWSTR=PWideChar',
    'LWSTDAPI=EXTERN_C DECLSPEC_IMPORT HRESULT STDAPICALLTYPE',
    'LWSTDAPIV=EXTERN_C DECLSPEC_IMPORT HRESULT STDAPIVCALLTYPE',
    'OPTIONAL=#',
    'OUT=#',
    'PASCAL=__stdcall',
    'PDWORD=DWORD *',
    'PSSTDAPI=STDAPI',
    'PURE=#',
    'PVOID=void *',
    'PWCHAR=wchar_t *',
    'SHDOCAPI=STDAPI',
    'SHFOLDERAPI=STDAPI',
    'SHSTDAPI=STDAPI',
    'SHSTDDOCAPI=STDAPI',
    'STDAPI=HRESULT STDAPICALLTYPE',
    'STDAPICALLTYPE=__stdcall',
    'STDAPIV=HRESULT STDAPIVCALLTYPE',
    'STDAPIVCALLTYPE=__cdecl',
    'STDCALL=__stdcall',
    'STDMETHODCALLTYPE=__stdcall',
    'THIS_=#',
    'virtual=#',
    'VOID=void',
    'WCHAR=wchar_t',
    'WINAPI=__stdcall',
    'WINAPIV=__cdecl',
    'WINOLEAPI=STDAPI',
    'WINOLEAUTAPI=STDAPI',
    'WINOLECTLAPI=STDAPI',
    'WINSHELLAPI=STDAPI',
    ''
  );

implementation

uses
  SysUtils;

{ TCParserData }

class procedure TCParserData.AddMacro(const AMacro: string);
begin
  if AMacro <> '' then
    FMacros.Add(AMacro);
end;

class procedure TCParserData.AddMacros(const Macros: array of string);
var
  I: Integer;
begin
  for I := 0 to High(Macros) do
    AddMacro(Macros[I]);
end;

class procedure TCParserData.AddType(const AType: string);
begin
  if AType <> '' then
    FTypes.Add(AType);
end;

class procedure TCParserData.AddTypes(const Types: array of string);
var
  I: Integer;
begin
  for I := 0 to High(Types) do
    AddType(Types[I]);
end;

class procedure TCParserData.Finalize;
begin
  FTypes.Free;
  FMacros.Free;
end;

class procedure TCParserData.Initialize;
begin
  FMacros := TStringList.Create;
  FMacros.CaseSensitive := True;
  FMacros.Duplicates := dupIgnore;
  AddMacros(CMacros);

  FTypes := TStringList.Create;
  FTypes.CaseSensitive := True;
  FTypes.Duplicates := dupIgnore;
  AddTypes(CTypes);
end;

class procedure TCParserData.SetTypes(const Types: TStrings);
begin
  FTypes.Clear;
  FTypes.AddStrings(Types);
end;

class function TCParserData.StringToType(S: string): TKnownType;
var
  I: Integer;
begin
  Result.PtrLevel := 0;
  for I := 1 to Length(S) do
    if S[I] = '*' then
      Inc(Result.PtrLevel);
  S := StringReplace(S, '*', '', [rfReplaceAll]);
  I := Pos('=', S);
  if I > 0 then
  begin
    Result.CType := Copy(S, 1, I - 1);
    Result.DelphiType := Copy(S, I + 1, Maxint);
  end;
end;

class procedure TCParserData.SetMacros(const Macros: TStrings);
begin
  FMacros.Clear;
  FMacros.AddStrings(Macros);
end;

class procedure TCParserData.SetOptions(const Value: TParserOptions);
begin
  FOptions := Value;
end;

initialization
  TCParserData.Initialize;

finalization
  TCParserData.Finalize;

end.


