{                                                                           }
{ File:       CParsers.pas                                                  }
{ Function:   Simple C parser for header conversion                         }
{ Language:   Delphi for Win32                                              }
{ Author:     Rudy Velthuis                                                 }
{ Copyright:  (c) 2008 Rudy Velthuis                                        }
{ Disclaimer: This code is freeware. All rights are reserved.               }
{             This code is provided as is, expressly without a warranty     }
{             of any kind. You use it at your own risk.                     }
{                                                                           }
{             If you use this code, please credit me.                       }
{                                                                           }

{ TODO: Review output mechanism -- it is too convoluted and clumsy. }
{ TODO: Perhaps the output should be delayed, when more data is collected.
        I am thinking of a real parse tree, i.e. the parameters of a
        function are a linked list of entries, just like the members of
        a struct. Comments and constant expressions of enumerators become
        part of the TDeclarationInfo, etc. }

unit CParsers;

interface

uses
  CLexers, SysUtils, Classes;

type
  TypeSpecifier = (tkNone, tkInt, tkChar, tkFloat, tkDouble, tkInt64,
                   tkVoid, tkWChar);
  SizeQualifier = (siNone, siShort, siLong, siInt64);
  SignQualifier = (sgNone, sgSigned, sgUnsigned);

//
//  TCParser is a simple recursive descent parser. The grammar I'm using
//  is most definitely not complete (a lot of C niceties are missing, but
//  these should be converted by the user, not by this simple tool anyway,
//  and a few corners were cut).
//
//  My own personal grammar notation (surely not correct, but good enough for a
//  simple converter):
//
//  (n)    group containing n
//  u | v  alternatives u and v
//  a?     zero or one a
//  b*     zero or more b
//  c+     one or more c
//  "x"    literal string x
//  z%     short for -- z ("," z)* -- i.e. one or more, comma-separated

// To be implemented:

//    function-typedef
//      "typedef" identifier "(" calling-convention "*" identifier ")" "("
//        (parameter-declaration%)* ")" ";" .

// Implemented:

//    function-decl:
//      return-specifier identifier "(" (parameter-declaration%)* ")" ";"  .
//
//    return-specifier:
//      extern-specifier? function-modifier* type-specifier function-modifier* .

{  hmmm... perhaps this is better:

      function-decl:
        return-specifier function-declarator "(" (parameter-declaration%)* ")" ";" .

      return-specifier:
        extern-specifier? function-modifier* type-specifier function-modifier* .

      function-declarator:
        pointer-specifier* identifier
}

//
//    function-modifier:
//      (calling-convention | declspec-modifier) .
//
//    calling-convention:
//      "__stdcall" | "_stdcall" | "__cdecl" | "_cdecl" .
//
//    declspec-modifier:
//      declspec-keyword "(" <token>+ ")" .
//
//    declspec-keyword:
//      "__declspec" | "_declspec"
//
//    extern-specifier:
//      "extern" ('"C"' | '"C++"') .
//
//    parameter-declaration:
//      parameter-specifier declarator .
//
//    parameter-specifier:
//      pointer-specifier* type-specifier .
//
//    define:
//      "#define" identifier ("(" identifier% ")")? constant-expression .
//
//    typedef:
//      "typedef" declaration declarator% ";" .
//
//    declaration-only:
//      declaration ";" .
//
//    declaration:
//      (struct_decl | union_decl | enum_decl | type-specification) .
//
//    struct-or-union-decl:
//      (("struct" | "union") identifier? "{" (declaration ";")+ "}") |
//      (("struct" | "union") identifier) .
//
//    enum-decl:
//      ("enum" identifier? "{" enumerator% ","? "}") |
//      ("enum" identifier) .
//
//    enumerator:
//      identifier ("=" constant-expression)? .
//
//    type-specification:
//      pointer-specifier* type-specifier .
//
//    declarator:
//      pointer-specifier* identifier array-specifier* .
//
//    type-specifier:
//      identifier |
//      (type-qualifier+ type-keyword? ) |
//      (type-qualifier* type-keyword) .
//
//    pointer-specifier:
//      ("*" | cv-qualifier)* .
//
//    array-specifier:
//      "[" constant-expression "]" .
//
//    type-qualifier:
//      "short" | "long" | "signed" | "unsigned" .
//
//    type-keyword:
//      "int" | "char" | "float" | "double" | "__int64" | "_int64" | "wchar_t" .
//
//    cv-qualifier:
//      "const" | "volatile" .
//
//  The functions quite closely follow the grammar. The name of each of them
//  is the grammar label prefixed with Parse and then camel capped.
//  So parameter-specifier is implemented in ParseParameterSpecifier.
//

  PDeclarationInfo = ^TDeclarationInfo;
  TDeclarationInfo = record
    TagName: string;            // tag of record, union or enum
    TypeName: string;           // type name
    Declarator: string;         // declarator name
    OldTypeName: string;        // type name before loop
    ArrayBounds: string;        // text for array bounds
    Keyword: string;            // "record", "union" or "enum"
    PrettyName: string;         // delphified typedef name
    PrettyTypeName: string;     // base type of delphified name
    PointerLevel: Integer;          // level of "*" pointers
    Convention: string;         // calling convention string
    Comment: string;            // comment belonging to this item
    Next: PDeclarationInfo;
  end;

  TCParser = class
  private
  var
    FLexer: TCLexer;
    FTokenType: TTokenType;
    FToken: string;
    FLastTokenType: TTokenType;
    FLastToken: string;
    FComment: string;
    FNestingLevel: Integer;
    FAutoEOL: Boolean;
    FCollectComments: Boolean;
    FError: string;
    FOutput: string;
    FOldToken: string;
  protected
    procedure ParseTrailingTypedefComment;
    procedure WriteFunctionDecl(const Info: TDeclarationInfo);
    procedure WriteComments;
    procedure FinishDeclaration(var Info: TDeclarationInfo);
    procedure DoPrettyName(var Info: TDeclarationInfo);
    procedure WriteSingleDeclName(var Info: TDeclarationInfo);
    procedure WriteStructuredDeclHeader(var Info: TDeclarationInfo);
    procedure ParseDeclaratorList(var Info: TDeclarationInfo);
    procedure PrettyNames(const Info: TDeclarationInfo);
    procedure ChooseDeclarationType(var Info: TDeclarationInfo);
    function Nesting: string;
    procedure WriteStr(const S: string);
    procedure WriteLine(const S: string = '');
    procedure WriteFront(const S: string);
    procedure WriteLineFront(const S: string = '');
    procedure Expected(const Msg: string);
    procedure MustMatch(const ExpectedToken: string);
    function Matches(const Token: string): Boolean;
    procedure Next;
    procedure ReplaceKnownTypes(var Info: TDeclarationInfo);
    function GetPrettyName(S: string; var PrettyName: string): Boolean;
    function ExternalSym(const Identifier: string): string;
    procedure FormatTypedef;
    procedure FormatFuncDecl;
    procedure WriteDeclaration(var Info: TDeclarationInfo);

    procedure ParseComment;
    function ParseEndOfLine: Boolean;
    function ParseCVQualifier: Boolean;
    function ParseTypeKeyword(var TypeKey: TypeSpecifier): Boolean;
    function ParseTypeQualifier(var Size: SizeQualifier; var Sign: SignQualifier): Boolean;
    function ParseParameterDeclaration: Boolean;
    function ParseReturnSpecifier(var Info: TDeclarationInfo): Boolean;
    function ParseFunctionTypedef: Boolean;
    function ParseFunctionModifier(var Info: TDeclarationInfo): Boolean;
    function ParseCallingConvention(var Info: TDeclarationInfo): Boolean;
    function ParseDeclspecModifier: Boolean;
    function ParseDeclspecKeyword: Boolean;
    function ParseExternSpecifier: Boolean;
    procedure ParseMacroDef;
    function ParseEnumerator: Boolean;
    procedure ParseMacroFunctionParams(const ParamType: string);
    procedure ParseMacroFunctionDecl(const MacroName: string);
    function ParsePointerSpecifier(var Info: TDeclarationInfo): Boolean;
    function ParseArraySpecifier(var Info: TDeclarationInfo): Boolean;
    function ParseTypeSpecifier(var Info: TDeclarationInfo): Boolean;
    function ParseDeclarator(var Info: TDeclarationInfo): Boolean;
    function ParseStructOrUnionDeclaration(var Info: TDeclarationInfo): Boolean;
    function ParseEnumDecl(var Info: TDeclarationInfo): Boolean;
    function ParseTypeSpecification(var Info: TDeclarationInfo): Boolean;
  public
    function ParseFunctionDeclaration: Boolean;
    function ParseDeclaration: Boolean;
    function ParseSimpleDeclaration: Boolean;
    procedure ParseTypedef;
    procedure ParseDefine;

    constructor Create(CharProvider: ICharProvider);

    property Error: string read FError;
    property Output: string read FOutput;
  end;

implementation

uses
{$IFNDEF UNICODE}
  UnicodeUtils,
{$ENDIF}
  CParserData;

{$IFDEF DEBUG}
const
{$ELSE}
resourcestring
{$ENDIF}
  SExpDeclaratorIdentifier = 'Declarator identifier expected, but %s found';
  SExpFinalFuncSemicolon = '; at end of function declaration';
  SExpFunctionName = 'function name';
  SExpMacroName = 'macro name';
  SExpEnumIdentifier = 'enumerator identifier. Found %s';

const
  CRLF = #13#10;

  CEscape = '&';
  CEnumStart = CEscape + '(';
  CEnumEnd = CEscape + ')';
  CommentStart = CEscape + '//';

  IntTypes: array[SizeQualifier, SignQualifier] of string = (
  // sgNone      sgSigned    sgUnsigned
    ('Integer',  'Integer',  'Cardinal'),  // siNone
    ('Smallint', 'Smallint', 'Word'),      // siShort
    ('Longint',  'Longint',  'Longword'),  // siLong
    ('Int64',    'Int64',    'UInt64')     // siInt64
  );

  CharTypes: array[SignQualifier] of string = (
  // sgNone      sgSigned    sgUnsigned
     'AnsiChar', 'Shortint', 'Byte'
  );

  Int64Types: array[SignQualifier] of string = (
  // sgNone      sgSigned    sgUnsigned
     'Int64',    'Int64',    'UInt64'
  );

  CDeclStr: array[Boolean] of string = (': ', ' = ');
  CPtrStr: array[Boolean] of string = ('P', '^');


{ -- Utility functions -- }

function LoCase(Ch: Char): Char;
begin
  if CharInSet(Ch, ['A'..'Z']) then
    Result := Char(Ord(Ch) - Ord('A') + Ord('a'))
  else
    Result := Ch;
end;

function IsUglyName(const S: string): Boolean;
var
  I: Integer;
begin
  Result := True;
  for I := 1 to Length(S) do
    if CharInSet(S[I], ['a'..'z']) then
    begin
      Result := False;
      Break;
    end;
end;

function TrimRightSpaces(const S: string): string;
var
  I: Integer;
begin
  Result := S;
  I := Length(Result);
  while Result[I] = ' ' do
    Dec(I);
  if I <> Length(Result) then
    SetLength(Result, I);
end;

{ -- TCParser -- }

//    function-typedef
//      "typedef" identifier "(" calling-convention "*" identifier ")" "("
//        (parameter-declaration%)* ")" ";" .
//
function TCParser.ParseFunctionTypedef: Boolean;
begin
  Result := True;
end;

//    function-decl:
//      return-specifier identifier "(" (parameter-declaration%)* ")" ";"  .
//
function TCParser.ParseFunctionDeclaration: Boolean;
var
  Info: TDeclarationInfo;
begin
  Info.TypeName := '';
  Info.Declarator := '';
  Info.ArrayBounds := '';
  Info.PointerLevel := 0;
  Info.Convention := '';

  ParseReturnSpecifier(Info);

  if FTokenType = ttIdentifier then
  begin
    Info.Declarator := FToken;
    Next;
  end
  else
    Expected(SExpFunctionName);

  MustMatch('(');

  FOldToken := FToken;

//  if Matches('void') and Matches(')') then
//    // nothing
//  else
  if Matches(')') then
    // nothing
  else
  begin
    FCollectComments := True;
    repeat
      ParseParameterDeclaration;
      Matches(',');
      WriteComments; // display any collected comments/EOLs.
    until Matches(')');
    FCollectComments := False;
  end;

  // Do not use Match(';') - this will call Next and thus scan over comments
  // and EOLs following the final ; and write them to output. It can also move
  // the CharProvider too far beyond the declaration.

  if FToken <> ';' then
    Expected(SExpFinalFuncSemicolon);

  WriteFunctionDecl(Info);
  Result := True;
end;

procedure TCParser.WriteFunctionDecl(const Info: TDeclarationInfo);
var
  FuncType: string;
begin
  if Info.TypeName <> '' then
    FuncType := 'function'
  else
    FuncType := 'procedure';

  WriteFront('{$EXTERNALSYM ' + Info.Declarator + '}' + CRLF +
    FuncType + ' ' + Info.Declarator + '(');
  FOutput := TrimRightSpaces(FOutput);
  WriteStr(')');
  if Info.TypeName <> '' then
    WriteStr(': ' + Info.TypeName);
  WriteStr('; ' + Info.Convention + ';');

  FormatFuncDecl;
end;

//    return-specifier:
//      extern-specifier function-modifier* type-specifier function-modifier* .
//
function TCParser.ParseReturnSpecifier(var Info: TDeclarationInfo): Boolean;
begin
  ParseExternSpecifier;
  ParseFunctionModifier(Info);
  ParseTypeSpecifier(Info);
  ParseFunctionModifier(Info);
  if Info.TypeName = 'void' then
    Info.TypeName := '';
  if Info.Convention = '' then
    Info.Convention := 'cdecl';
  Result := True;
end;

//    function-modifier:
//      (calling-convention | declspec-modifier) .
//
function TCParser.ParseFunctionModifier(var Info: TDeclarationInfo): Boolean;
begin
  Result := False;
  if ParseCallingConvention(Info) then
    Result := True;
  if ParseDeclspecModifier then
    Result := True;
  if ParseCallingConvention(Info) then
    Result := True;
end;

//    calling-convention:
//      "__stdcall" | "_stdcall" | "__cdecl" | "_cdecl" .
//
function TCParser.ParseCallingConvention(var Info: TDeclarationInfo): Boolean;
begin
  Result := True;
  if Matches('__stdcall') or Matches('_stdcall') then
    Info.Convention := 'winapi'
  else if Matches('__cdecl') or Matches('_cdecl') then
    Info.Convention := 'cdecl'
  else
    Result := False;
end;

//    declspec-modifier:
//      (declspec-keyword "(" <token>+ ")")? .
//
function TCParser.ParseDeclspecModifier: Boolean;
var
  ParenLevel: Integer;
begin
  Result := ParseDeclspecKeyword;
  if Result then
  begin
    ParenLevel := 0;
    if Matches('(') then
    begin
      Inc(ParenLevel);
      while ParenLevel > 0 do
      begin
        Next;
        if Matches('(') then
          Inc(ParenLevel)
        else if Matches(')') then
          Dec(ParenLevel);
      end;
    end;
  end;
end;

//    declspec-keyword:
//      "__declspec" | "_declspec" .
//
function TCParser.ParseDeclspecKeyword: Boolean;
begin
  Result := Matches('__declspec') or Matches('_declspec');
end;

//    extern-specifier:
//      ("extern" ('"C"' | '"C++"'))? .
//
function TCParser.ParseExternSpecifier: Boolean;
begin
  Result := Matches('extern');
  if Result then
  begin
    if FTokenType = ttStringLiteral then
      Next;
  end;
end;

//    parameter-declaration:
//      type-specification declarator .
//
function TCParser.ParseParameterDeclaration: Boolean;
var
  Info: TDeclarationInfo;
begin
  Info.Declarator := '';
  Info.TypeName := '';
  Info.ArrayBounds := '';
  Info.PointerLevel := 0;

  ParseTypeSpecification(Info);
  ParseDeclarator(Info);

  if Info.ArrayBounds <> '' then
    Inc(Info.PointerLevel);

  if (Info.PointerLevel >= 1) and
     ((Info.TypeName = 'AnsiChar') or (Info.TypeName = 'Byte')) then
  begin
    Dec(Info.PointerLevel);
    Info.TypeName := 'P' + Info.TypeName;
  end;

  ReplaceKnownTypes(Info);

  { TODO: make this optional }
  if Info.PointerLevel > 0 then
  begin
    WriteStr('var ');
    Dec(Info.PointerLevel);
  end;

  WriteStr(Info.Declarator + ': ' +
    StringOfChar('P', Info.PointerLevel) + Info.TypeName);
  if FToken = ',' then
    WriteStr('; ');
  Result := True;
end;

// Special general declaration parser for swap declaration.
function TCParser.ParseSimpleDeclaration: Boolean;
var
  Info: TDeclarationInfo;
begin
  Info.Declarator := '';
  Info.TypeName := '';
  Info.ArrayBounds := '';
  Info.PointerLevel := 0;
  FCollectComments := True;

  ParseTypeSpecification(Info);
  ParseDeclarator(Info);

  if (Info.PointerLevel > 0) and
     ((Info.TypeName = 'AnsiChar') or (Info.TypeName = 'Byte')) then
  begin
    Dec(Info.PointerLevel);
    Info.TypeName := 'P' + Info.TypeName;
  end;

  ReplaceKnownTypes(Info);

  WriteStr(Info.Declarator + ': ');
  if Info.ArrayBounds <> '' then
    WriteStr('array [' + Info.ArrayBounds + '] of ');
  WriteStr(StringOfChar('P', Info.PointerLevel) + Info.TypeName);
  if (FToken = ',') or (FToken = ';') then
    WriteStr('; ');

  Result := True;
end;

//  define:
//    "#define" identifier("(" identifier% ")")? constant-expression .
//
procedure TCParser.ParseDefine;
var
  MacroName: string;
begin
  MustMatch('#');
  MustMatch('define');

  if FTokenType = ttIdentifier then
  begin
    MacroName := FToken;
    Next;
  end
  else
    Expected(SExpMacroName);

  // It is only a macro function if there is no space between macro name
  // and leading parenthesis. Otherwise the parenthesis is just part of the
  // replacement text.
  if not CharInSet(FLexer.LastChar, [#0..#31, ' ']) and Matches('(') then
    ParseMacroFunctionDecl(MacroName)
  else
  begin
    // This assumes - most of the time correctly - that the macro is a simple
    // true constant or type alias declaration.
    WriteLine('  {$EXTERNALSYM ' + MacroName + '}');
    WriteStr('  ' + MacroName + StringOfChar(' ', 36 - Length(MacroName)) + '= ');
    ParseMacroDef;
  end;
end;

(* TODO: structs, unions and enums with tag names but no following declarators
        fail, at the moment. They are possible and should work. The tag name
        should become the declarator.

        IOW, this does not parse:

          enum enumSQLSCMCommands {SQLSCMCmd_STOP,
                                   SQLSCMCmd_PAUSE,
                                   SQLSCMCmd_START,
                                   SQLSCMCmd_CONTINUE};
*)

// NEW syntax, to be implemented -
//  typedef:
//    ("typedef" declaration declarator% ";") | (declaration ";") .

// OLD syntax -
//  typedef:
//    "typedef" declaration ";" .
//
procedure TCParser.ParseTypedef;
//var
//  Info: TDeclarationInfo;}
begin
  Matches('typedef');
  ParseDeclaration;

//  if Matches('typedef') then
//  begin
//    ParseDeclaration(Info);
//    ParseDeclaratorList(Info);
//    PrettyNames(Info);
//  end;

  FAutoEOL := False;
  Matches(';');
  ParseTrailingTypedefComment;
end;

//  declaration:
//    (struct_decl | union_decl | enum_decl | type-specification) declarator% .
//
function TCParser.ParseDeclaration: Boolean;
var
  Info: TDeclarationInfo;
  SaveOutput: string;
begin
  SaveOutput := FOutput;
  FOutput := '';
  
  ChooseDeclarationType(Info);

  if Info.TagName <> '' then
    Info.TypeName := Info.TagName;

  Info.OldTypeName := Info.TypeName;

  ParseDeclaratorList(Info);
  PrettyNames(Info);

  WriteFront(SaveOutput);
  FormatTypeDef;
  Result := True;
end;

//  struct-or-union-decl:
//    (("struct" | "union") identifier? "{" (declaration ";")+ "}") |
//    (("struct" | "union") identifier) .
//
function TCParser.ParseStructOrUnionDeclaration(var Info: TDeclarationInfo): Boolean;
begin

  Next; // read past "struct" or "union"

  if FTokenType = ttIdentifier then
  begin
    Info.TagName := FToken;
    Next;
  end;

  if FToken <> '{' then
  begin
    Info.TypeName := Info.TagName;
    Info.TagName := '';
    Info.Keyword := ''; // prevent "end;" from being inserted somewhere.
  end
  else
  begin
    Next; // read past "{"
    Inc(FNestingLevel);
    repeat
      ParseDeclaration;
      MustMatch(';');
    until Matches('}');
    Dec(FNestingLevel);
  end;
  Result := True;
end;

//  enum-decl:
//    ("enum" identifier? "{" enumerator% ","? "}") |
//    ("enum" identifier) .
//
function TCParser.ParseEnumDecl(var Info: TDeclarationInfo): Boolean;
begin
  Next; // read past "enum"

  if FTokenType = ttIdentifier then
  begin
    Info.TagName := FToken;
    Next;
  end;

  if FToken <> '{' then
  begin
    Info.TypeName := Info.TagName;
    Info.TagName := '';
  end
  else
  begin
    Next; // read past "{"
    if FLastTokenType = ttEOL then
      WriteLineFront;
    Inc(FNestingLevel, 2);
    repeat
      ParseEnumerator;
//      Matches(',');
    until Matches('}');
    Dec(FNestingLevel, 2);
  end;

  Result := True;
end;

//  enumerator:
//    identifier ("=" constant-expression)? .
//
function TCParser.ParseEnumerator: Boolean;
//var
//  Comment: string;
begin
  Result := True;
//  Comment := FComment;

  if FTokenType = ttIdentifier then
  begin
    WriteStr(FToken + ' ');
    Next;
  end
  else
    Expected(Format(SExpEnumIdentifier, [FToken]));

  if Matches('=') then
  begin
    FOutput := TrimRightSpaces(FOutput);
    WriteStr(' = ');
    FCollectComments := True;
    while (FToken <> ',') and (FToken <> '}') do
    begin
      if FTokenType = ttOperator then
        WriteStr(' ' + TCLexer.ConvertOperator(FToken) + ' ')
      else if FToken = '(' then
        WriteStr('(')
      else if FToken = ')' then
      begin
        FOutput := TrimRightSpaces(FOutput);
        WriteStr(')');
      end
      else
        WriteStr(FToken + ' ');
      Next;
    end;
    FCollectComments := False;
    WriteComments;
  end;

  // do not use Matches(','), since that can call Next, which calls Comments
  // and EndOfLine before the comma can be written to Output.
  if FToken = ',' then
  begin
    FOutput := TrimRightSpaces(FOutput);
    WriteStr(', ')
  end;


  Matches(',');
end;

//  type-specification:
//    pointer-specifier* type-specifier .
//
function TCParser.ParseTypeSpecification(var Info: TDeclarationInfo): Boolean;
begin
  repeat until not ParsePointerSpecifier(Info);
  Result := ParseTypeSpecifier(Info);
end;

//  declarator:
//    pointer-specifier* identifier array-specifier* .
//
function TCParser.ParseDeclarator(var Info: TDeclarationInfo): Boolean;
begin
  repeat until not ParsePointerSpecifier(Info);
  if FTokenType = ttIdentifier then
  begin
    Info.Declarator := FToken;
    Next;
  end
  else
  if (FOldToken = 'void') and (Info.PointerLevel = 0) then
  begin
    Result := True;
    Exit;
  end
  else
    raise Exception.CreateFmt(SExpDeclaratorIdentifier, [FToken]);
  while FToken = '[' do
  begin
    Next;
    ParseArraySpecifier(Info);
    MustMatch(']');
  end;
  Result := True;
end;

//  type-specifier:
//    identifier |
//    (type-qualifier+ type-keyword? ) |
//    (type-qualifier* type-keyword) .
//
function TCParser.ParseTypeSpecifier(var Info: TDeclarationInfo): Boolean;
var
  Size: SizeQualifier;
  Sign: SignQualifier;
  TypeKey: TypeSpecifier;
begin
  Result := True;
  if FTokenType = ttIdentifier then
  begin
    Info.TypeName := FToken;
    Next;
  end
  else
  begin
    Size := siNone;
    Sign := sgNone;

    // Let's assume valid syntax
    repeat until not ParseTypeQualifier(Size, Sign);
    if not ParseTypeKeyword(TypeKey) then
      TypeKey := tkInt;

    case TypeKey of
      tkInt:
        Info.TypeName := IntTypes[Size, Sign];
      tkChar:
        Info.TypeName := CharTypes[Sign];
      tkFloat:
        Info.TypeName := 'Single';
      tkDouble:
        if Size = siLong then
          Info.TypeName := 'Extended'
        else
          Info.TypeName := 'Double';
      tkInt64:
        Info.TypeName := Int64Types[Sign];
      tkVoid:
        Info.TypeName := 'void';
      tkWChar:
        Info.TypeName := 'WideChar';
      else
        Result := False;
    end;
  end;
end;

//  pointer-specifier:
//    ("*" | cv-qualifier)* .
//
function TCParser.ParsePointerSpecifier(var Info: TDeclarationInfo): Boolean;
begin
  if ParseCVQualifier then
    Result := True
  else if (FToken = '*') or (FToken = '&') then
  begin
    Result := True;
    Inc(Info.PointerLevel);
    Next;
  end
  else
    Result := False;
end;

// This could be replaced by a full expression parser, but that is, IMO
// not really a necessity. If there are a few mistakes, like using div where /
// is required, the compiler will catch it and warn the user.
// Also, since operators like &= or ++ can't be translated, I'll leave the
// invalid names, so the compiler can catch them.
// IOW, this is good enough for my purposes.

//  array-specifier:
//    "[" constant-expression "]" .
//
function TCParser.ParseArraySpecifier(var Info: TDeclarationInfo): Boolean;
var
  UpperBound: string;
  UpperValue: Integer;
begin
  UpperBound := '';
  Result := True;

  while FToken <> ']' do
  begin
    if FTokentype = ttOperator then
      UpperBound := UpperBound + FLexer.ConvertOperator(FToken) + ' '
    else
      UpperBound := UpperBound + FToken + ' ';
    Next;
  end;
  UpperBound := Trim(UpperBound);

  if TryStrToInt(UpperBound, UpperValue) then
    UpperBound := IntToStr(UpperValue - 1)
  else
    UpperBound := UpperBound + ' - 1';

  if Info.ArrayBounds <> '' then
    Info.ArrayBounds := Info.ArrayBounds + ', ';
  Info.ArrayBounds := Info.ArrayBounds + '0..' + UpperBound;

end;

//  type-qualifier:
//    "short" | "long" | "signed" | "unsigned" .
//
function TCParser.ParseTypeQualifier(var Size: SizeQualifier;
  var Sign: SignQualifier): Boolean;
begin
  Result := True;

  if FToken = 'short' then
    Size := siShort
  else if FToken = 'long' then
    if Size = siLong then
      Size := siInt64 // "long long"
    else
      Size := siLong
  else if FToken = 'signed' then
    Sign := sgSigned
  else if FToken = 'unsigned' then
    Sign := sgUnsigned
  else
    Result := False;
  if Result then
    Next;
end;

//  type-keyword:
//    "int" | "char" | "float" | "double" | "__int64" | "_int64" | "wchar_t" .
//
function TCParser.ParseTypeKeyword(var TypeKey: TypeSpecifier): Boolean;
begin
  Result := True;
  TypeKey := tkNone;
  if FToken = 'int' then
    TypeKey := tkInt
  else if FToken = 'char' then
    TypeKey := tkChar
  else if FToken = 'float' then
    TypeKey := tkFloat
  else if FToken = 'double' then
    TypeKey := tkDouble
  else if (FToken = '__int64') or (FToken = '_int64') then
    TypeKey := tkInt64
  else if FToken = 'void' then
    TypeKey := tkVoid
  else if FToken = 'wchar_t' then
    TypeKey := tkWChar
  else
    Result := False;
  if Result then
    Next;
end;

//  cv-qualifier:
//    "const" | "volatile" .
//
function TCParser.ParseCVQualifier: Boolean;
begin
  Result := (FToken = 'const') or (FToken = 'volatile');
  if Result then
    Next;
end;

procedure TCParser.Expected(const Msg: string);
begin
  FError := FError + 'Expected ' + Msg + CRLF;
  raise Exception.Create(FError);
end;

function TCParser.ExternalSym(const Identifier: string): string;
begin
  if FNestingLevel = 0 then
    Result := '{$EXTERNALSYM ' + Identifier + '}' + CRLF + Nesting
  else
    Result := '';
end;

// TODO: somehow, in records and unions, singe line comments do not get
//       concatenated with their code lines.

procedure TCParser.FormatFuncDecl;
var
  I, P: Integer;
  List: TStringList;
  MaxCommentPos: Integer;
begin
  List := TStringList.Create;
  try
    // Kludge
    FOutput := StringReplace(FOutput, '(: void', '(', []);
    List.Text := FOutput;

    MaxCommentPos := 0;

    // First pass: remove empty lines and find comment positions.
    for I := List.Count - 1 downto 0 do
    begin
      List[I] := TrimRight(List[I]);

      if List[I] = '' then
        List.Delete(I)
      else
      begin
        P := Pos(CEscape, List[I]);
        if P > MaxCommentPos then
          MaxCommentPos := P;
      end;
    end;

    // Second pass: align comments, remove escapes
    for I := 0 to List.Count - 1 do
    begin
      P := Pos(CEscape, List[I]);
      if P > 0 then
        List[I] := StringReplace(List[I], CEscape, StringOfChar(' ', MaxCommentPos - P), []);
    end;

    FOutput := StringReplace(List.Text, CEscape, '', [rfReplaceAll]);
  finally
    List.Free;
  end;
end;

procedure TCParser.FormatTypedef;
var
  List: TStringList;
  NestingLevel: Integer;
  I: Integer;
  P: Integer;
  InEnum: Boolean;
  MaxEqualsPos: Integer;
  MaxCommentPos: Integer;
begin
  List := TStringList.Create;
  try
    List.Text := FOutput;

    // First pass: remove empty lines and concatenate comments.
    for I := List.Count - 1 downto 0 do
    begin
      List[I] := Trim(List[I]);

      P := Pos(CommentStart, List[I]);
      if P <= 0 then
        P := Pos('&{', List[I]);
      if P = 1 then
      begin
        if I > 0 then
        begin
          List[I - 1] := List[I - 1] + ' ' +
                         StringOfChar(' ', 33 - Length(List[I - 1])) +
                         List[I];
          List.Delete(I);
        end;
      end
      else
        if List[I] = '' then
          List.Delete(I);
    end;

    // Second pass: nesting
    Nestinglevel := 0;
    MaxEqualsPos := 0;
    InEnum := False;
    for I := 0 to List.Count - 1 do
    begin
      if Pos('end;', List[I]) > 0 then
        Dec(NestingLevel);
      if (Pos(CEnumEnd, List[I]) > 0) and (Pos(CEnumStart, List[I]) <= 0) then
      begin
        InEnum := False;
        Dec(NestingLevel);
      end;
      List[I] := StringOfChar(' ', 2 * NestingLevel + 2) + List[I] + ' ';
      if InEnum then
      begin
        P := Pos('=', List[I]);
        if P > MaxEqualsPos then
          MaxEqualsPos := P;
      end;
      if (Pos(CEnumStart, List[I]) > 0) and (Pos(CEnumEnd, List[I]) <= 0) then
      begin
        InEnum := True;
        Inc(NestingLevel);
      end;
      if (Pos('record', List[I]) > 0) or (Pos('union', List[I]) > 0) then
        Inc(NestingLevel);
    end;

    // Third pass: align '=' --> StringReplace parts in last pass
    InEnum := False;
    MaxCommentPos := 0;
    for I := 0 to List.Count - 1 do
    begin
      if Pos(CEnumStart, List[I]) > 0 then
        InEnum := True;
      if Pos(CEnumEnd, List[I]) > 0 then
        InEnum := False;
      if InEnum then
      begin
        P := Pos('=', List[I]);
        if (P > 0) and (P < MaxEqualsPos) then
          List[I] := StringReplace(List[I], '=', StringOfChar(' ', MaxEqualsPos - P) + '=', []);
      end;
      P := Pos('//', List[I]);
      if P > MaxCommentPos then
        MaxCommentPos := P;
      P := Pos('{', List[I]);
      if (Pos('{$', List[I]) <> P) and (P > MaxCommentPos) then
        MaxCommentPos := P;
    end;

    // Fourth pass: align comments
    for I := 0 to List.Count - 1 do
    begin
      P := Pos('{', List[I]);
      if (P <> Pos('{$', List[I])) and (P < MaxCommentPos) then
        List[I] := StringReplace(List[I], '{', StringOfChar(' ', MaxCommentPos - P) + '{', []);
      P := Pos('//', List[I]);
      if P < MaxCommentPos then
        List[I] := StringReplace(List[I], '//', StringOfChar(' ', MaxCommentPos - P) + '//', []);
      List[I] := StringReplace(List[I], CEscape, '', [rfReplaceAll]);
    end;

    FOutput := List.Text;
  finally
    List.Free;
  end;
end;

procedure TCParser.MustMatch(const ExpectedToken: string);
begin
  if ExpectedToken <> FToken then
    Expected(ExpectedToken)
  else
    Next;
end;

function TCParser.Matches(const Token: string): Boolean;
begin
  Result := FToken = Token;
  if Result then
    Next;
end;

procedure TCParser.Next;
begin
  FLastTokenType := FTokenType;
  FLastToken := FToken;
  FTokenType := FLexer.NextToken(FToken);

  if FAutoEOL then
  begin
//    {!!} WriteStr('|' + FToken + '| ');
    ParseComment;
    ParseEndOfLine;
  end;
end;

constructor TCParser.Create(CharProvider: ICharProvider);
begin
  FNestingLevel := 0;
  FLexer := TCLexer.Create(CharProvider);
  FTokenType := ttEOL;
  FToken := #13;
  FAutoEOL := True;
  FCollectComments := False;
  Next;
end;

procedure TCParser.WriteComments;
begin
  if FComment <> '' then
    WriteStr(FComment);
  FComment := '';
end;

procedure TCParser.ParseMacroDef;
begin
  FAutoEOL := False;
  try
    while not (FTokenType in [ttEOL, ttEOF]) do
    begin
      if FTokenType = ttOperator then
        WriteStr(FLexer.ConvertOperator(FToken) + ' ')
      else if FToken = '(' then
        WriteStr(FToken)
      else if (FToken = ')') or (FToken = ',') then
      begin
        FOutput := TrimRightSpaces(FOutput);
        WriteStr(FToken);
        if FToken = ',' then
          WriteStr(' ');
      end
      else if FTokenType in [ttSingleComment, ttMultiComment] then
        FComment := FComment + ' ' + FToken
      else
        WriteStr(FToken + ' ');
      Next;
    end;
  finally
    FAutoEOL := True;
  end;
  FOutput := TrimRightSpaces(FOutput);
  WriteLine('; ' + FComment);
end;

procedure TCParser.ParseMacroFunctionParams(const ParamType: string);
begin
  while not Matches(')') do
  begin
    if FTokenType = ttIdentifier then
    begin
      WriteStr(FToken + ': ' + ParamType);
      Next;
    end;
    if Matches(',') then
      WriteStr('; ');
  end;
end;

procedure TCParser.ParseMacroFunctionDecl(const MacroName: string);
var
  ParamType: string;
begin
  if poDWords in TCParserData.Options then
    ParamType := 'DWORD'
  else
    ParamType := '';
  WriteLine('{$EXTERNALSYM ' + MacroName + '}');
  WriteStr('function ' + MacroName + '(');
  ParseMacroFunctionParams(ParamType);
  WriteLine('): ' + ParamType + ';');
  if poInlines in TCParserData.Options then
    WriteLine('  inline;');
  WriteLine('(*');
  WriteLine('begin');
  WriteStr('  Result := ');
  ParseMacroDef;
  WriteLine('end;');
  WriteLine('*)');
end;

procedure TCParser.WriteDeclaration(var Info: TDeclarationInfo);
begin
  // The body of structured types is already written.
  if Info.Keyword = 'enum' then
    Info.Keyword := CEnumStart;

  WriteStructuredDeclHeader(Info);
  FinishDeclaration(Info);
  ReplaceKnownTypes(Info);

  if (Info.TypeName = '') and (Info.TagName = '') then
  begin
    Info.OldTypeName := Info.Declarator;
    Info.TypeName := Info.Declarator;
  end;

  WriteSingleDeclName(Info);
  DoPrettyName(Info);
end;

function TCParser.ParseEndOfLine: Boolean;
begin
  Result := FTokenType = ttEOL;
  if Result then
  begin
    if FCollectComments then
      FComment := FComment + CRLF + Nesting
    else
    begin
      if FComment <> '' then
      begin
        WriteLine(); // to prevent problems with formatting - I don't particularly like this solution
        WriteStr(Trim(FComment));
        FComment := '';
      end;
      WriteLine();
      WriteStr(Nesting);
    end;
    Next;
  end;
end;

procedure TCParser.ParseComment;
begin
  if FTokenType in [ttMultiComment, ttSingleComment] then
  begin
    if FLastTokenType <> ttEOL then
      FComment := FComment + ' ' + CEscape + FToken
    else
      FComment := FComment + ' ' + FToken;
    Next;
  end;
end;

function TCParser.Nesting: string;
begin
  Result := StringOfChar(' ', FNestingLevel * 2 + 2);
end;

procedure TCParser.WriteStr(const S: string);
begin
  FOutput := FOutput + S;
end;

procedure TCParser.WriteFront(const S: string);
begin
  FOutput := S + FOutput;
end;

procedure TCParser.WriteLine(const S: string);
begin
  FOutput := FOutput + S + CRLF;
end;

procedure TCParser.WriteLineFront(const S: string);
begin
  FOutput := S + CRLF + FOutput;
end;

function TCParser.GetPrettyName(S: string; var PrettyName: string): Boolean;
type
  TState = (Lower, Upper);
var
  I: Integer;
  IsUgly: Boolean;
  State: TState;
begin
  S := Trim(S);
  IsUgly := IsUglyName(S);
  PrettyName := '';
  State := Upper;
  for I := 1 to Length(S) do
    if S[I] = '_' then
      State := Upper
    else
      if State = Upper then
      begin
        PrettyName := PrettyName + UpCase(S[I]);
        State := Lower;
      end
      else
      begin
        if IsUgly then
          PrettyName := PrettyName + LoCase(S[I])
        else
          PrettyName := PrettyName + S[I];
      end;
  Result := True;
end;

procedure TCParser.ReplaceKnownTypes(var Info: TDeclarationInfo);
var
  Found: string;
  Level: Integer;
begin
  Level := Info.PointerLevel;
  Found := TCParserData.Types.Values[Info.TypeName + StringOfChar('*', Level)];
  while (Level >= 0) and (Found = '') do
  begin
    Dec(Level);
    Found := TCParserData.Types.Values[Info.TypeName + StringOfChar('*', Level)];
  end;
  if Found <> '' then
  begin
    Info.TypeName := Found;
    Dec(Info.PointerLevel, Level);
  end;
end;

procedure TCParser.WriteSingleDeclName(var Info: TDeclarationInfo);
begin
  if (Info.Declarator <> '') and (Info.TypeName <> Info.Declarator) then
  begin
    if FNestingLevel = 0 then
      WriteStr(ExternalSym(Info.Declarator));
    WriteStr(Info.Declarator + CDeclStr[FNestingLevel = 0]);
    if Info.ArrayBounds <> '' then
      WriteStr('array[' + Info.ArrayBounds + '] of ');
    if Info.PointerLevel > 0 then
      WriteStr(CPtrStr[FNestinglevel = 0] + StringOfChar('P', Info.PointerLevel - 1));
    WriteLine(Info.TypeName + ';');
    WriteStr(Nesting);
  end;
end;

procedure TCParser.WriteStructuredDeclHeader(var Info: TDeclarationInfo);
var
  Name: string;
  S: string;
begin
  Name := '';
  if (Info.Keyword <> '') then
  begin
    if (Info.PointerLevel = 0) and (Info.ArrayBounds = '') and (Info.TagName = '') then
      Name := Info.Declarator
    else if Info.TagName <> '' then
      Name := Info.TagName;
    if Name <> '' then
    begin
      S := ExternalSym(Name) + Name + CDeclStr[FNestingLevel = 0] + Info.Keyword;
      if Info.Keyword <> CEnumStart then
        S := S + CRLF;
      WriteFront(S);
    end;
  end;
end;

procedure TCParser.ParseDeclaratorList(var Info: TDeclarationInfo);
begin
  repeat
    Info.PointerLevel := 0;
    Info.TypeName := Info.OldTypeName;
    Info.Declarator := '';
    Info.ArrayBounds := '';
    ParseDeclarator(Info);
    WriteDeclaration(Info);
  until not Matches(',');
end;

procedure TCParser.PrettyNames(const Info: TDeclarationInfo);
begin
  if Info.PrettyName <> '' then
  begin
    WriteFront('P' + Info.PrettyName + ' = ^T' + Info.PrettyName + ';' + CRLF);
    WriteLine('T' + Info.PrettyName + ' = ' + Info.PrettyTypeName + ';');
  end;
end;

procedure TCParser.ChooseDeclarationType(var Info: TDeclarationInfo);
begin
  if (FToken = 'struct') then
  begin
    Info.Keyword := 'record';
    ParseStructOrUnionDeclaration(Info);
  end
  else if FToken = 'union' then
  begin
    Info.Keyword := 'union';
    ParseStructOrUnionDeclaration(Info);
  end
  else if FToken = 'enum' then
  begin
    Info.Keyword := 'enum';
    ParseEnumDecl(Info);
  end
  else
  begin
    ParseTypeSpecification(Info);
    Info.Keyword := '';
  end;
end;

procedure TCParser.FinishDeclaration(var Info: TDeclarationInfo);
begin
  if (Info.Keyword = 'record') or (Info.Keyword = 'union') then
  begin
    WriteLine('end; ');
    Info.Keyword := '';
  end
  else if Info.Keyword = CEnumStart then
  begin
    FOutput := TrimRightSpaces(FOutput);
    WriteLine(CEnumEnd + '; ');
    Info.Keyword := '';
  end;
end;

procedure TCParser.DoPrettyName(var Info: TDeclarationInfo);
begin
  if (Info.PrettyName = '') and
     (FNestingLevel = 0) and
     (Info.PointerLevel = 0) and
     (Info.ArrayBounds = '') then
  begin
    GetPrettyName(Info.Declarator, Info.PrettyName);
    Info.PrettyTypeName := Info.TypeName;
  end;
end;

{ TDeclarationInfo }


procedure TCParser.ParseTrailingTypedefComment;
begin
  while not (FTokenType in [ttEOL, ttEOF]) do
  begin
    if FTokenType in [ttSingleComment, ttMultiComment] then
      FComment := FComment + ' ' + FToken;
    FTokenType := FLexer.NextToken(FToken);
  end;
  if FComment <> '' then
    WriteLine(' ' + FComment);
end;

end.

