program ParserTest;

{$APPTYPE CONSOLE}

uses
  SysUtils,
  TypInfo,
  CLexers in '..\Common\CLexers.pas',
  CParsers in '..\Common\CParsers.pas',
  CParserData in '..\Common\CParserData.pas',
  CParserUtils in '..\Common\CParserUtils.pas';

type
  TCharProvider = class(TInterfacedObject, ICharProvider)
  private
    Text: string;
    Index: Integer;
  public
    constructor Create(const S: string);
    function NextChar: Char;
    function LookAhead: Char;
    function CurrentPosition: Integer;
    function StartPosition: Integer;
    function FullText: string;
  end;

{ TCharProvider }

constructor TCharProvider.Create(const S: string);
begin
  Text := S;
  Index := 1;
end;

function TCharProvider.CurrentPosition: Integer;
begin
  Result := Index;
end;

function TCharProvider.FullText: string;
begin
  Result := Copy(Text, 1, Index - 1);
end;

function TCharProvider.LookAhead: Char;
begin
  if Index <= Length(Text) then
    Result := Text[Index]
  else
    Result := CLexers.LexerEndMarker;
end;

function TCharProvider.NextChar: Char;
begin
  if Index > Length(Text) then
  begin
    Result := LexerEndMarker;
    Index := Length(Text) + 2;
  end
  else
  begin
    Result := Text[Index];
    Inc(Index);
  end;
end;

function TCharProvider.StartPosition: Integer;
begin
  Result := 1;
end;

function StringOf(C: Char): string;
begin
  if CharInSet(C, [#0..#31, #127..#255]) then
    Result := '#' + IntToStr(Ord(C))
  else
    Result := '''' + C + '''';
end;

procedure TestLexer(const prog: string);
var
  CurrentToken: string;
  TokenType: TTokenType;
  Lexer: TCLexer;
begin
  Lexer := TCLexer.Create(TCharProvider.Create(prog));
  try
    Writeln(prog);
    Writeln;
    repeat
      TokenType := Lexer.NextToken(CurrentToken);
      begin
        Writeln(GetEnumName(TypeInfo(TTokenType), Ord(TokenType)):20, ' ', CurrentToken:30,
          ' Last: ', StringOf(Lexer.LastChar), ' Cur: ', StringOf(Lexer.CurrentChar),
          ' Ahead: ', StringOf(Lexer.LookAhead));
      end;
    until TokenType = ttEOF;
  finally
    Lexer.Free;
  end;
  Writeln;
  Writeln;
  Readln;
end;

procedure TestCParser(const prog: string);
var
  Parser: TCParser;
  CharProvider: ICharProvider;
begin
  CharProvider := TCharProvider.Create(prog);
  Parser := TCParser.Create(CharProvider);
  try
    Writeln;
    Writeln(StringOfChar('_', 78));
    Writeln(prog);
    Writeln(StringOfChar('-', 78));
    Writeln;

    try
      Parser.ParseTypedef;
    except
      Writeln(Parser.Error);
    end;

    Writeln(Parser.Output);
  finally
    Parser.Free;
  end;
end;

procedure TestDefine(const prog: string);
var
  Parser: TCParser;
begin
  Parser := TCParser.Create(TCharProvider.Create(prog));
  try
    Writeln;
    Writeln(StringofChar('=', 78));
    Writeln(prog);
    Writeln(StringOfChar('-', 78));
    Writeln;
    try
      Parser.ParseDefine;
    except
      Writeln(Parser.Error);
    end;
    Writeln(Parser.Output);
  finally
    Parser.Free;
  end;
end;

procedure TestFunc(const prog: string);
var
  Parser: TCParser;
  Provider: ICharProvider;
begin
  Provider := TCharProvider.Create(prog);
  Parser := TCParser.Create(Provider);
  try
    Writeln;
    Writeln(StringOfChar('=', 78));
    Writeln(prog);
    Writeln(StringOfChar('-', 78));
    Writeln;

    try
      Parser.ParseFunctionDeclaration;
    except
      Writeln(Parser.Error);
    end;
    Writeln(Provider.FullText);
    Writeln(StringOfChar('_', 78));
    Writeln(Parser.Output);
  finally
    Parser.Free;
  end;
end;

begin
  try
//    TestLexer('#define IDO_SHGIOI_SHARE  0x0FFF1234'#13#10'hello this is it');
//    TestLexer('#define MyConst /* hello */ (a) // there');
//    TestLexer('typedef enum Bla{one=L"Hi",two=271,three="Hello",four=0,five=0x0a0123,six=0376,}BLA,*LPBLA;');
//    TestLexer('typedef struct Rudy{intz a;int b[maxint + 4];char d;float *f[0];}RUDY,*LPRUDY;');
    TestLexer('HRESULT WINAPI Bla(int a,void *c,LPVOID qrs/*** this is a comment ***/,char *f[],float q);');
    TestLexer('LWSTDAPIV Bla(int i[], ...);');
    TestLexer('extern "C" __declspec(dllimport) HRESULT DDRNew(char *sArchitecture, char *sAgentName, char *sSiteCode);');
    TestLexer('typedef struct _SHELLEXECUTEINFOW'#13#10 +
      '{'#13#10 +
      '        DWORD cbSize;'#13#10 +
      '        ULONG fMask;'#13#10 +
      '        HWND hwnd;'#13#10 +
      '        LPCWSTR  lpVerb;'#13#10 +
      '        LPCWSTR  lpFile;'#13#10 +
      '        LPCWSTR  lpParameters;'#13#10 +
      '        LPCWSTR  lpDirectory;'#13#10 +
      '        int nShow;'#13#10 +
      '        HINSTANCE hInstApp; // single line comment'#13#10 +
      '        // Optional fields'#13#10 +
      '        LPVOID lpIDList;'#13#10 +
      '        LPCWSTR  lpClass;'#13#10 +
      '        HKEY hkeyClass[16];'#13#10 +
      '        DWORD dwHotKey;'#13#10 +
      '        union {'#13#10 +
      '            HANDLE hIcon;'#13#10 +
      '            HANDLE hMonitor;'#13#10 +
      '        } DUMMYUNIONNAME;'#13#10 +
      '        HANDLE hProcess;'#13#10 +
      '} SHELLEXECUTEINFOW, *LPSHELLEXECUTEINFOW;');
    TestLexer('#define ListView_GetItemCount(a, b) sendmessage(0, 12, 33L, 0L, 21u, 33, 44);'#13#10);
    TestCParser('typedef int bla');
    TestCParser('typedef const int * const *bla[11][13];');
    TestCParser('typedef const LONG FAR * FAR * const willy[MAXINT + 4];');
    TestCParser('typedef const SHELLEXECUTEINFOW **lplprudy, *lprudy, rudy, arudy[17], * const crudy;');
    TestCParser('typedef const unsigned long weird, *lpweird, **lplpweird; /* this is a comment */');
    TestCParser('typedef     ULONG   ACCESS_RIGHTS, *PACCESS_RIGHTS;');
    TestCParser('typedef PVOID CLFS_LOG_ARCHIVE_CONTEXT, *PCLFS_LOG_ARCHIVE_CONTEXT; //hello');
    TestCParser('typedef void *CLFS_LOG_ARCHIVE_CONTEXT, **PCLFS_LOG_ARCHIVE_CONTEXT;');
    TestCParser('typedef FILE *PFILE, **PPFILE;');
    TestCParser('typedef char *argv[1];');
    TestCParser('typedef char **argv;');
    TestCParser('typedef const GUID *pguid;');
    TestCParser('typedef const char _FAR *__s;');
    TestCParser('typedef struct Silly *LPSilly; // hello');
    TestCParser('typedef FILETIME const *pft1;');
    TestCParser('typedef ResourceString const * const rsrc;');
    TestCParser('typedef WCHAR const * const *papwszInsertionText;');
    TestCParser('typedef struct tag'#13#10 +
                '{'#13#10 +
                '    int a; //hello'#13#10 +
                '    double b;'#13#10 +
                '    int *c;'#13#10 +
                '    unsigned int d[12]; //array'#13#10 +
                '} test, *ptest;');
    TestCParser('struct daisy *PDAISY;');
    TestCParser('typedef enum TagEnum { a, b, c, d } name, *lpname; // hello');
    TestCParser('typedef enum TagEnum {'#13#10 +
      '    a = 1, // comment'#13#10 +
      '    b, // comment'#13#10 +
      '    california, // comment'#13#10 +
      '    d, // comment'#13#10 +
      '} TypeName, *LPTypeName;'#13#10 +
      '// comment below'#13#10 +
      '// more comments'#13#10 +
      'this is just text');
    TestCParser('typedef enum TagEnum { a = 1, b, california, d } TypeName, *LPTypeName;'#13#10 +
      '// comment below'#13#10 +
      '// more comments'#13#10 +
      'this is just text');
    TestCParser('typedef enum TagEnum {'#13#10 +
      '    a,'#13#10 +
      '    b,'#13#10 +
      '    california,'#13#10 +
      '    d,'#13#10 +
      '} TypeName, *LPTypeName;'#13#10 +
      '// comment below'#13#10 +
      '// more comments'#13#10 +
      'this is just text');
    TestCParser('typedef enum _tagAUTOCOMPLETELISTOPTIONS'#13#10 +
      '{'#13#10 +
      '    ACLO_NONE            = 0,    // don''t enumerate anything'#13#10 +
      '    ACLO_CURRENTDIR      = 1,    // enumerate current directory'#13#10 +
      '    ACLO_MYCOMPUTER      = 2,    // enumerate MyComputer'#13#10 +
      '    ACLO_DESKTOP         = 4,    // enumerate Desktop Folder'#13#10 +
      '    ACLO_FAVORITES       = 8,    // enumerate Favorites Folder'#13#10 +
      '    ACLO_FILESYSONLY     = 16,   // enumerate only the file system'#13#10 +
      '    ACLO_FILESYSDIRS     = 32,   // enumerate only the file system dirs, UNC shares, and UNC servers.'#13#10 +
      '    ACLO_VIRTUALNAMESPACE = 64,  // enumereate on the virual namespace'#13#10 +
      '} AUTOCOMPLETELISTOPTIONS;');
    TestCParser('typedef enum _tagAUTOCOMPLETELISTOPTIONS'#13#10 +
      '{'#13#10 +
      '    ACLO_NONE            = 0,    // don''t enumerate anything'#13#10 +
      '    ACLO_CURRENTDIR      = 1,    // enumerate current directory'#13#10 +
      '    ACLO_MYCOMPUTER      = 2,    // enumerate MyComputer'#13#10 +
      '    ACLO_DESKTOP         = 4,    // enumerate Desktop Folder'#13#10 +
      '    ACLO_FAVORITES       = 8,    // enumerate Favorites Folder'#13#10 +
      '    ACLO_FILESYSONLY     = 16,   // enumerate only the file system'#13#10 +
      '    ACLO_FILESYSDIRS     = 32,   // enumerate only the file system dirs, UNC shares, and UNC servers.'#13#10 +
      '    ACLO_VIRTUALNAMESPACE = 64,  // enumereate on the virual namespace'#13#10 +
      '} AUTOCOMPLETELISTOPTIONS;');
    TestCParser('enum _tagAUTOCOMPLETELISTOPTIONS'#13#10 +
      '{'#13#10 +
      '    ACLO_NONE            = 0,    // don''t enumerate anything'#13#10 +
      '    ACLO_CURRENTDIR      = 1,    // enumerate current directory'#13#10 +
      '    ACLO_MYCOMPUTER      = 2,    // enumerate MyComputer'#13#10 +
      '    ACLO_DESKTOP         = 4,    // enumerate Desktop Folder'#13#10 +
      '    ACLO_FAVORITES       = 8,    // enumerate Favorites Folder'#13#10 +
      '    ACLO_FILESYSONLY     = 16,   // enumerate only the file system'#13#10 +
      '    ACLO_FILESYSDIRS     = 32,   // enumerate only the file system dirs, UNC shares, and UNC servers.'#13#10 +
      '    ACLO_VIRTUALNAMESPACE = 64  // enumereate on the virual namespace'#13#10 +
      '} AUTOCOMPLETELISTOPTIONS;');
    TestCParser('typedef'#13#10 +
      'enum APO_FLAG'#13#10 +
      '    {'#9'APO_FLAG_NONE	= 0,'#13#10 +
      #9'APO_FLAG_INPLACE'#9'= 0x1,'#13#10 +
      #9'APO_FLAG_SAMPLESPERFRAME_MUST_MATCH	= 0x2, // demo'#13#10 +
      #9'APO_FLAG_FRAMESPERSECOND_MUST_MATCH	= 0x4,'#13#10 +
      #9'APO_FLAG_BITSPERSAMPLE_MUST_MATCH	= 0x8,'#13#10 +
      #9'APO_FLAG_DEFAULT	= ( ( APO_FLAG_SAMPLESPERFRAME_MUST_MATCH | APO_FLAG_FRAMESPERSECOND_MUST_MATCH )  | APO_FLAG_BITSPERSAMPLE_MUST_MATCH )'#13#10 +
      '    } '#9'APO_FLAG;');
    TestCParser('typedef struct _SHELLEXECUTEINFOW'#13#10 +
      '{'#13#10 +
      '        DWORD cbSize;'#13#10 +
      '        ULONG *fMask;'#13#10 +
      '        HWND hwnd;'#13#10 +
      '        LPCWSTR  lpVerb;'#13#10 +
      '        LPCWSTR  lpFile;'#13#10 +
      '        LPCWSTR  lpParameters;'#13#10 +
      '        LPCWSTR  lpDirectory;'#13#10 +
      '        int nShow;'#13#10 +
      '        HINSTANCE hInstApp; // single line comment'#13#10 +
      '        // Optional fields'#13#10 +
      '        LPVOID lpIDList;'#13#10 +
      '        LPCWSTR  lpClass; /* multiline comment'#13#10 +
      '                             ends here */'#13#10 +
      '        HKEY hkeyClass[15][3];'#13#10 +
      '        DWORD dwHotKey;'#13#10 +
      '/*'#13#10 +
      ' * This is a multiline comment in front of the u-nion'#13#10 +
      ' * to see what happens to it'#13#10 +
      ' */'#13#10 +
      '        union {'#13#10 +
      '          HANDLE hIcon;'#13#10 +
      '          HANDLE hMonitor;'#13#10 +
      '        } DUMMYUNIONNAME;'#13#10 +
      '        HANDLE hProcess;'#13#10 +
      '} SHELLEXECUTEINFOW, *LPSHELLEXECUTEINFOW; // test'#13#10'{ This is still inked to it }'#13#10'int a');
    TestCParser('typedef struct Rudy{intz a;int b[maxint + 4];char d;float *f[0];}RUDY,*LPRUDY;');
    TestCParser('typedef /* [wire_marshal] */ const ITEMIDLIST __unaligned *LPCITEMIDLIST;');
    TestCParser('typedef struct _SHITEMID'#13#10 +
      '    {'#13#10 +
      '    USHORT cb;'#13#10 +
      '    BYTE abID[ 1 ];'#13#10 +
      '    } 	SHITEMID;');
    TestCParser('typedef struct _ITEMIDLIST'#13#10 +
      '    {'#13#10 +
      '    SHITEMID mkid;'#13#10 +
      '    } 	ITEMIDLIST;');
    TestCParser('typedef struct _STRRET'#13#10 +
      '    {'#13#10 +
      '    UINT uType;'#13#10 +
      '    union'#13#10 +
      '        {'#13#10 +
      '        LPWSTR pOleStr;'#13#10 +
      '        UINT uOffset;'#13#10 +
      '        char cStr[ 260 ];'#13#10 +
      '        } 	DUMMYUNIONNAME;'#13#10 +
      '    } 	STRRET;');
    TestCParser('typedef struct'#13#10 +
      '{'#13#10 +
      '    DWORD           dwSize;'#13#10 +
      '    DWORD           dwMask;              // IN/OUT  Which Attributes to Get/Set'#13#10 +
      '    SHELLVIEWID*    pvid;                // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      '    // The folder''s WebView template path'#13#10 +
      '    LPWSTR          pszWebViewTemplate;   // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      '    DWORD           cchWebViewTemplate;   // IN - Specifies the size of the buffer pointed to by pszWebViewTemplate'#13#10 +
      '                                         // Ignored if dwReadWrite is FCS_READ'#13#10 +
      '    LPWSTR           pszWebViewTemplateVersion;  // currently IN only'#13#10 +
      '    // Infotip for the folder'#13#10 +
      '    LPWSTR          pszInfoTip;          // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      '    DWORD           cchInfoTip;          // IN - Specifies the size of the buffer pointed to by pszInfoTip'#13#10 +
      '                                         // Ignored if dwReadWrite is FCS_READ'#13#10 +
      '    // CLSID that points to more info in the registry'#13#10 +
      '    CLSID*          pclsid;              // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      '    // Other flags for the folder. Takes FCS_FLAG_* values'#13#10 +
      '    DWORD           dwFlags;             // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      ''#13#10 +
      ''#13#10 +
      '    LPWSTR           pszIconFile;        // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      '    DWORD            cchIconFile;        // IN - Specifies the size of the buffer pointed to by pszIconFile'#13#10 +
      '                                         // Ignored if dwReadWrite is FCS_READ'#13#10 +
      ''#13#10 +
      '    int              iIconIndex;         // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      ''#13#10 +
      '    LPWSTR           pszLogo;        // OUT - if dwReadWrite is FCS_READ, IN - otherwise'#13#10 +
      '    DWORD            cchLogo;        // IN - Specifies the size of the buffer pointed to by pszIconFile'#13#10 +
      '                                         // Ignored if dwReadWrite is FCS_READ'#13#10 +
      '} SHFOLDERCUSTOMSETTINGS, *LPSHFOLDERCUSTOMSETTINGS;');
    TestCParser('typedef struct _CERT_POLICY_QUALIFIER_INFO {'#13#10 +
      '    LPSTR                       pszPolicyQualifierId;   // pszObjId */'#13#10 +
      '    CRYPT_OBJID_BLOB            Qualifier;              // optional */'#13#10 +
      '} CERT_POLICY_QUALIFIER_INFO, *PCERT_POLICY_QUALIFIER_INFO;'#13#10);
    TestCParser('typedef struct _CERT_POLICY_QUALIFIER_INFO {'#13#10 +
      '    LPSTR                       pszPolicyQualifierId;'#13#10 +
      '    CRYPT_OBJID_BLOB            Qualifier;              '#13#10 +
      '} CERT_POLICY_QUALIFIER_INFO, *PCERT_POLICY_QUALIFIER_INFO;'#13#10);
    TestCParser('typedef struct _CERT_POLICY_QUALIFIER_INFO {'#13#10 +
      '    LPSTR                       pszPolicyQualifierId;   // pszObjId'#13#10 +
      '    CRYPT_OBJID_BLOB            Qualifier;              // optional'#13#10 +
      '} CERT_POLICY_QUALIFIER_INFO, *PCERT_POLICY_QUALIFIER_INFO;'#13#10 +
      #13#10 +
      '//this is a test'#13#10 +
      'typedef struct _CERT_POLICY_INFO {'#13#10 +
      '    LPSTR                       pszPolicyIdentifier;    // pszObjId'#13#10 +
      '    DWORD                       cPolicyQualifier;       // optional'#13#10 +
      '    CERT_POLICY_QUALIFIER_INFO  *rgPolicyQualifier;'#13#10 +
      '} CERT_POLICY_INFO, *PCERT_POLICY_INFO;'#13#10 +
      #13#10 +
      '//this is a test'#13#10 +
      'typedef struct _CERT_POLICIES_INFO {'#13#10 +
      '    DWORD                       cPolicyInfo;'#13#10 +
      '    CERT_POLICY_INFO            *rgPolicyInfo;'#13#10 +
      '} CERT_POLICIES_INFO, *PCERT_POLICIES_INFO;'#13#10 +
      '// certenrolls_end'#13#10);
    TestCParser('typedef struct _VTableProvStruc {'#13#10 +
      '    DWORD   Version;'#13#10 +
      '    FARPROC FuncVerifyImage;'#13#10 +
      '    FARPROC FuncReturnhWnd;'#13#10 +
      '    DWORD   dwProvType;'#13#10 +
      '    BYTE        *pbContextInfo;'#13#10 +
      '    DWORD       cbContextInfo;'#13#10 +
      '    LPSTR   pszProvName;'#13#10 +
      '} VTableProvStruc, *PVTableProvStruc;'#13#10 +
      '#endif //(NTDDI_VERSION < NTDDI_WINXP)'#13#10);
    TestCParser('typedef struct _CRYPT_RC4_KEY_STATE {'#13#10 +
      '    unsigned char Key[16];'#13#10 +
      '    unsigned char SBox[256];'#13#10 +
      '    unsigned char i;'#13#10 +
      '    unsigned char *j;'#13#10 +
      '} CRYPT_RC4_KEY_STATE, *PCRYPT_RC4_KEY_STATE;'#13#10#13#10);
    TestDefine('#define Anton Geesink // hallo anton');
    TestDefine('#define Aard Schenk');
    TestDefine('#define macrofunc(a, b) { x = 3 * a + b; }');
    TestDefine('#define MyConst L"Hello"');
    TestDefine('#define MyConst (x)');
    TestDefine('#define MyConst (a | b & ~ c)');
    TestDefine('#define MyConst (a|b&~c)');
    TestDefine('#define MyConst (a>>+3,b>>=5,a>>3|b>>5)');
    TestDefine('#define MyConst (a >>+ 3, b >>= 5,a >> 3|b >> 5)');
    TestDefine('#define MyConst /* hello */ (a) // there');
    TestDefine('#define SIOM_OVERLAYINDEX         1');
    TestDefine('#define IDO_SHGIOI_SHARE  0x0FFF1234'#13#10'hello this is it');
    TestDefine('#define SHGetIconOverlayIndex  SHGetIconOverlayIndexW');
    TestDefine('#define EXP_SZ_LINK_SIG                0xA0000001   // LPEXP_SZ_LINK (target)');
    TestDefine('#define EXP_SZ_LINK_SIG                0xA0000001');
    TestDefine('#define CSIDL_DESKTOP                   0x0000        // <desktop>');
    TestDefine('#define CSIDL_FLAG_MASK                 0xFF00        // mask for all possible flag values');
    TestDefine('#define DIRID_SYSTEM            11              // system32');
    TestDefine('#define SPFILENOTIFY_QUEUESCAN          0x0000000f');
    TestDefine('#define STATUS_SEVERITY(hr)  (((hr) >> 30) & 0x3) //hello?'#13#10'#define my ass');
    TestDefine('#define my ass');
    TestDefine('#define MAXULONG_PTR (~((ULONG_PTR)0))');
    TestDefine('#define MAXLONG_PTR  ((LONG_PTR)(MAXULONG_PTR >> 1))');
    TestDefine('#define UlongToPtr(ul) ULongToPtr(ul)');
    TestDefine('#define GET_NAP(_bth_addr)  ((USHORT) (((_bth_addr) & NAP_MASK) >> NAP_BIT_OFFSET))'#13#10 +
      '#define GET_NAP(_bth_addr)  ((USHORT) (((_bth_addr) & NAP_MASK) >> NAP_BIT_OFFSET)');
    TestDefine('#define    PixelFormat1bppIndexed     (1 | ( 1 << 8) | PixelFormatIndexed | PixelFormatGDI)');
    TestDefine('#define GET_ALG_CLASS(x)                (x & (7 << 13))'#13#10 +
      '#define GET_ALG_TYPE(x)                 (x & (15 << 9))'#13#10);
    TestDefine('#define ListView_GetItemCount(hwnd) (int)SendMessage((hwnd), LVM_GETITEMCOUNT, 0, 0L)'#13#10);
    TestDefine('#define ListView_GetItemCount(a, b) sendmessage(0, 12, 33L, 0L, 21u, 33, 44)'#13#10);
    TestFunc('extern "C" int __stdcall __declspec(dllimport) bla(int a);');
    TestFunc('HRESULT WINAPI foo(int *b);');
    TestFunc('STDAPI novo(int a, int b, char *s);');
    TestFunc('HRESULT WINAPI YesOrNo(int *yes, char *msg);');
    TestFunc('HRESULT WINAPI Toot(BOOL *yes,'#13#10'    char *msg);');
    TestFunc('DECLSPEC_IMPORT'#13#10 +
      'BOOL'#13#10 +
      'WINAPI'#13#10 +
      'CertComparePublicKeyInfo('#13#10 +
      '    DWORD dwCertEncodingType,'#13#10 +
      '    PCERT_PUBLIC_KEY_INFO pPublicKey1,'#13#10 +
      '    PCERT_PUBLIC_KEY_INFO pPublicKey2'#13#10 +
      '    );');
    TestFunc('DECLSPEC_IMPORT'#13#10 +
      'BOOL'#13#10 +
      'WINAPI'#13#10 +
      'CryptGetOIDFunctionAddress('#13#10 +
      '    HCRYPTOIDFUNCSET hFuncSet, // one'#13#10 +
      '    DWORD dwEncodingType, // two'#13#10 +
      '    LPCSTR pszOID, // three */'#13#10 +
      '    DWORD dwFlags, // four'#13#10 +
      '    void **ppvFuncAddr, // five */'#13#10 +
      '    HCRYPTOIDFUNCADDR *phFuncAddr // six */'#13#10 +
      ');'#13#10 +
      '    '#13#10);
    TestFunc('void Bla(int a /* test */ );');
    TestFunc('DECLSPEC_IMPORT'#13#10 +
      'BOOL'#13#10 +
      'WINAPI'#13#10 +
      'CryptHashData('#13#10 +
      '  HCRYPTHASH  hHash,'#13#10 +
      '  const BYTE  *pbData,'#13#10 +
      '  DWORD   dwDataLen,'#13#10 +
      '  DWORD   dwFlags'#13#10 +
      '    );'#13#10);
    TestFunc('DECLSPEC_IMPORT'#13#10 +
      'BOOL'#13#10 +
      'WINAPI'#13#10 +
      'CryptQueryObject('#13#10 +
      '   DWORD                    dwObjectType,'#13#10 +
      '   const void               *pvObject,'#13#10 +
      '   DWORD                    dwExpectedContentTypeFlags,'#13#10 +
      '   DWORD                    dwExpectedFormatTypeFlags,'#13#10 +
      '   DWORD                    dwFlags,'#13#10 +
      '   DWORD               *pdwMsgAndCertEncodingType,'#13#10 +
      '   DWORD               *pdwContentType,'#13#10 +
      '   DWORD               *pdwFormatType,'#13#10 +
      '   HCERTSTORE          *phCertStore,'#13#10 +
      '   HCRYPTMSG           *phMsg,'#13#10 +
      '   const void **ppvContext'#13#10 +
      '    );'#13#10 +
      ''#13#10 +
      ''#13#10 +
      '//-------------------------------------------------------------------------'#13#10 +
      '//dwObjectType for CryptQueryObject'#13#10 +
      '//-------------------------------------------------------------------------'#13#10 +
      '#define     CERT_QUERY_OBJECT_FILE         0x00000001'#13#10 +
      '#define     CERT_QUERY_OBJECT_BLOB         0x00000002'#13#10);
    Readln;
  except
    on E: Exception do
      Writeln(E.Classname, ': ', E.Message);
  end;
end.

